#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Schema processor."""

import json
import logging
from typing import Tuple
import voluptuous as vol

from .const import (
    ATTR_CONTROLLER,
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE_HTG,
    ATTR_DHW_VALVE,
    ATTR_HTG_CONTROL,
    ATTR_ORPHANS,
    ATTR_STORED_HOTWATER,
    ATTR_SYSTEM,
    ATTR_UFH_CONTROLLERS,
    ATTR_ZONE_TYPE,
    ATTR_ZONE_SENSOR,
    ATTR_ZONES,
    DEFAULT_MAX_ZONES,
    ZONE_TYPE_SLUGS,
    __dev_mode__,
    id_to_address as addr,
)

# false = False; null = None; true = True

# TODO: duplicated in __init__.py
DONT_CREATE_MESSAGES = 3
DONT_CREATE_ENTITIES = 2
DONT_UPDATE_ENTITIES = 1

DEVICE_ID_REGEXP = r"^[0-9]{2}:[0-9]{6}$"
DEVICE_ID = vol.Match(DEVICE_ID_REGEXP)

DOMAIN_ID_REGEXP = r"^[0-9A-F]{2}$"
DOMAIN_ID = vol.Match(DOMAIN_ID_REGEXP)

ZONE_IDX_REGEXP = r"^0[0-9AB]$"  # TODO: what if > 12 zones? (e.g. hometronics)
ZONE_IDX = vol.Match(ZONE_IDX_REGEXP)

ZONE_SCHEMA = vol.Schema(
    {
        vol.Required(ZONE_IDX): vol.Any(
            None,
            {
                vol.Optional(ATTR_ZONE_TYPE, default=None): vol.Any(
                    None, vol.Any(*list(ZONE_TYPE_SLUGS))
                ),
                vol.Optional(ATTR_ZONE_SENSOR, default=None): vol.Any(None, DEVICE_ID),
                vol.Optional(ATTR_DEVICES, default=[]): vol.Any(None, [DEVICE_ID]),
            },
        )
    }
)
SER2NET_SCHEMA = vol.Schema(
    {vol.Required("enabled"): bool, vol.Optional("socket", default="0.0.0.0:5000"): str}
)

CONFIG_SCHEMA = vol.Schema(
    {
        vol.Optional("disable_sending", default=False): vol.Any(None, bool),
        vol.Optional("disable_discovery", default=False): vol.Any(None, bool),
        vol.Optional("enforce_allowlist", default=False): vol.Any(None, bool),
        vol.Optional("enforce_blocklist", default=True): vol.Any(None, bool),
        vol.Optional("evofw_flag", default=None): vol.Any(None, bool),
        # vol.Optional("input_file"): vol.Any(None, str),
        vol.Optional("max_zones", default=DEFAULT_MAX_ZONES): vol.Any(None, int),
        vol.Optional("packet_log", default=None): vol.Any(None, str),
        vol.Optional("reduce_processing", default=0): vol.Any(None, int),
        vol.Optional("serial_port"): vol.Any(None, str),
        vol.Optional("ser2net_relay"): SER2NET_SCHEMA,
        vol.Optional("use_schema", default=True): vol.Any(None, bool),
    },
    extra=vol.ALLOW_EXTRA,
)
DHW_SCHEMA = vol.Schema(
    {
        vol.Optional(ATTR_DHW_SENSOR): vol.Any(None, vol.Match(r"^07:[0-9]{6}$")),
        vol.Optional(ATTR_DHW_VALVE): vol.Any(None, vol.Match(r"^13:[0-9]{6}$")),
        vol.Optional(ATTR_DHW_VALVE_HTG): vol.Any(None, vol.Match(r"^13:[0-9]{6}$")),
    }
)
UFH_SCHEMA = vol.Schema({})
ORPHAN_SCHEMA = vol.Schema(
    {vol.Optional(ATTR_ORPHANS): vol.Any(None, vol.All(DEVICE_ID))}
)
SYSTEM_SCHEMA = vol.Schema(
    {
        vol.Required(ATTR_CONTROLLER): vol.Match(r"^(01|23):[0-9]{6}$"),
        vol.Optional(ATTR_SYSTEM, default={}): vol.Schema(
            {
                vol.Optional(ATTR_HTG_CONTROL): vol.Any(
                    None, vol.Match(r"^(10|13):[0-9]{6}$")
                ),
                vol.Optional(ATTR_ORPHANS): vol.Any(None, vol.All([DEVICE_ID])),
            }
        ),
        vol.Optional(ATTR_STORED_HOTWATER): vol.Any(None, DHW_SCHEMA),
        vol.Optional(ATTR_ZONES): vol.Any(
            None, vol.All(ZONE_SCHEMA, vol.Length(min=1, max=12))
        ),
        vol.Optional(ATTR_UFH_CONTROLLERS): vol.Any(None, [UFH_SCHEMA]),
    },
    extra=vol.ALLOW_EXTRA,
)
# GLOBAL_SCHEMA = vol.Schema(
#     vol.Any(SYSTEM_SCHEMA, vol.Length(min=0)), ORPHAN_SCHEMA, extra=vol.ALLOW_EXTRA
# )
KNOWNS_SCHEMA = vol.Schema(
    {
        vol.Optional(DEVICE_ID): vol.Any(
            None,
            {
                vol.Optional("name", default=None): vol.Any(None, str),
                vol.Optional("_parent_zone"): vol.Any(None, DOMAIN_ID),
                vol.Optional("_has_battery"): vol.Any(None, bool),
            },
        )  # , extra=vol.ALLOW_EXTRA
    }
)
FILTER_SCHEMA = vol.Schema(
    {
        vol.Optional("allowlist"): vol.Any(None, vol.All(KNOWNS_SCHEMA)),
        vol.Optional("blocklist"): vol.Any(None, vol.All(KNOWNS_SCHEMA)),
    }
)
# SCHEMA = vol.Schema(
#     {
#         vol.Optional("configuration"): CONFIG_SCHEMA,
#         vol.Optional("global_schema"): GLOBAL_SCHEMA,
#         vol.Optional("known_devices"): KNOWNS_SCHEMA,
#     }
# )
MONITOR_SCHEMA = vol.Schema(
    {
        vol.Optional("probe_system"): vol.Any(None, str),
        vol.Optional("execut_cmd"): vol.Any(None, str),
        vol.Optional("evofw_flag"): vol.Any(None, str),
        vol.Optional("packet_log"): vol.Any(None, str),
    }
)
PARSE_SCHEMA = vol.Schema({})
CLI_SCHEMA = vol.Schema({})

_LOGGER = logging.getLogger(__name__)
if __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def load_config(gwy, **config) -> Tuple[dict, list, list]:
    """Process the schema, and the configuration and return True if it is valid."""

    # def proc_cli(config):
    # self.config["input_file"] = config.get("input_file")
    # config["known_devices"] = config.get("known_devices")
    # self.config["reduce_processing"] = config.get("raw_output", 0)

    # if self.serial_port and self.config["input_file"]:
    #     _LOGGER.warning(
    #         "Serial port specified (%s), so ignoring input file (%s)",
    #         self.serial_port,
    #         self.config["input_file"],
    #     )
    #     self.config["input_file"] = None

    # self.config["disable_sending"] = not config.get("probe_system")
    # if self.config["input_file"]:
    #     self.config["disable_sending"] = True

    # if self.config["reduce_processing"] >= DONT_CREATE_MESSAGES:
    #     config["message_log"] = None
    #     _stream = (None, sys.stdout)
    # else:
    #     _stream = (sys.stdout, None)

    schema_filename = config.get("config_file")

    if schema_filename is None:
        return {}, (), ()

    try:
        with open(schema_filename) as schema_fp:
            config = json.load(schema_fp)
    except FileNotFoundError:  # if it doesn't exist, create it later
        return {}, (), ()

    # config = SCHEMA(config)
    params, schema = config["configuration"], config["global_schema"]
    gwy.known_devices = config["known_devices"]

    allowlist = list(config["known_devices"]["allowlist"])
    blocklist = list(config["known_devices"]["blocklist"])

    if params["use_allowlist"] and allowlist:
        _list = True
    elif params["use_blocklist"] and blocklist:
        _list = False
    elif params["use_allowlist"] is not False and allowlist:
        _list = True
    elif params["use_blocklist"] is not False and blocklist:
        _list = False
    else:
        _list = None

    if params["use_schema"]:  # regardless of filters, & updates known/allow
        (load_schema(gwy, k, v) for k, v in schema.items() if k != ATTR_ORPHANS)

    if _list:
        allowlist += [d for d in gwy.device_by_id if d not in allowlist]
        blocklist = []
    elif _list is False:
        allowlist = []
        blocklist = [d for d in blocklist if d not in gwy.device_by_id]
    else:
        allowlist = blocklist = []  # cheeky, but OK

    return params, tuple(allowlist), tuple(blocklist)


def load_schema(gwy, schema, **kwargs) -> dict:
    """Process the schema, and the configuration and return True if it is valid."""
    # TODO: check a sensor is not a device in another zone

    if schema.get(ATTR_CONTROLLER) is None:
        return {}

    schema = SYSTEM_SCHEMA(schema)

    ctl_id = schema[ATTR_CONTROLLER]
    ctl = gwy.get_device(addr(ctl_id), controller=addr(ctl_id))

    gwy.evo = ctl

    if ATTR_HTG_CONTROL in schema[ATTR_SYSTEM]:
        htg_id = schema[ATTR_SYSTEM][ATTR_HTG_CONTROL]
        if htg_id:
            ctl.boiler_control = gwy.get_device(addr(htg_id), controller=ctl)

    if ATTR_STORED_HOTWATER in schema:
        dhw = schema[ATTR_STORED_HOTWATER]
        if dhw:
            ctl.dhw = ctl.get_zone("FA")

            dhw_sensor_id = dhw.get(ATTR_DHW_SENSOR)
            if dhw_sensor_id is not None:
                ctl.dhw._set_sensor(gwy.get_device(addr(dhw_sensor_id), controller=ctl))

            dhw_id = dhw.get(ATTR_DHW_VALVE)
            if dhw_id is not None:
                ctl.dhw.hotwater_valve = gwy.get_device(addr(dhw_id), controller=ctl)

            htg_id = dhw.get(ATTR_DHW_VALVE_HTG)
            if htg_id is not None:
                ctl.dhw.heating_valve = gwy.get_device(addr(htg_id), controller=ctl)

    if schema.get(ATTR_ZONES):
        for zone_idx, attr in schema[ATTR_ZONES].items():
            zone = ctl.get_zone(zone_idx, zone_type=attr.get(ATTR_ZONE_TYPE))

            sensor_id = attr.get(ATTR_ZONE_SENSOR)
            if sensor_id:
                zone._set_sensor(gwy.get_device(addr(sensor_id), controller=ctl))

            device_list = attr.get(ATTR_DEVICES)
            if device_list:
                for device_id in attr.get(ATTR_DEVICES):
                    gwy.get_device(addr(device_id), controller=ctl, domain_id=zone_idx)

    # for ufh_ctl, ufh_schema in schema[ATTR_UFH_CONTROLLERS]:
    #     dev = gwy.get_device(addr(ufh_ctl), controller=ctl)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""The evohome-compatible zones."""

from abc import ABCMeta, abstractmethod
import asyncio
import json
import logging
from typing import Optional

from .command import Schedule
from .const import (
    ATTR_DEVICES,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE,
    ATTR_DHW_VALVE_HTG,
    ATTR_HEAT_DEMAND,
    ATTR_SETPOINT,
    ATTR_TEMP,
    ATTR_OPEN_WINDOW,
    ATTR_ZONE_SENSOR,
    ATTR_ZONE_TYPE,
    # CODE_000C_DEVICE_TYPE,
    DEVICE_HAS_ZONE_SENSOR,
    DHW_STATE_MAP,
    DISCOVER_SCHEMA,
    DISCOVER_PARAMS,
    DISCOVER_STATUS,
    DISCOVER_ALL,
    ZONE_CLASS_MAP,
    ZONE_TYPE_MAP,
    ZONE_TYPE_SLUGS,
    ZONE_MODE_LOOKUP,
    ZONE_MODE_MAP,
    __dev_mode__,
)
from .devices import Device, Entity, HeatDemand, _dtm
from .exceptions import CorruptStateError

_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


def _temp(value) -> str:
    """Return a two's complement Temperature/Setpoint."""
    if value is None:
        return "7FFF"

    try:
        value = float(value)
    except ValueError:
        raise ValueError("Invalid temperature")

    if value < 0:
        raise ValueError("Invalid temperature")

    return f"{int(value*100):04X}"


class ZoneBase(Entity, metaclass=ABCMeta):
    """The Zone/DHW base class."""

    def __init__(self, controller, zone_idx) -> None:
        _LOGGER.debug(
            "Creating a Domain: %s_%s %s", controller.id, zone_idx, self.__class__
        )
        super().__init__(controller._gwy, controller=controller)
        assert zone_idx not in controller.zone_by_idx, "Duplicate zone idx"

        self.id = f"{controller.id}_{zone_idx}"
        self.idx = zone_idx

        self._name = None
        self._zone_type = None

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return json.dumps(self.schema, indent=2)

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""
        return f"{self.id} ({self._zone_type})"

    @abstractmethod
    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        raise NotImplementedError

    def _send_cmd(self, code, **kwargs) -> None:
        dest = kwargs.pop("dest_addr", self._ctl.id)
        payload = kwargs.pop("payload", f"{self.idx}00")
        super()._send_cmd(code, dest, payload, **kwargs)

    @property
    @abstractmethod
    def schema(self) -> dict:
        """Return the schema of the zone/DHW (e.g. sensor_id, zone_type)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def params(self) -> dict:
        """Return the configuration of the zone/DHW (e.g. min_temp, overrun)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def status(self) -> dict:
        """Return the current state of the zone/DHW (e.g. setpoint, temperature)."""
        raise NotImplementedError

    @property
    @abstractmethod
    def sensor(self) -> Device:
        """Return the temperature sensor of the zone/DHW."""
        raise NotImplementedError

    @abstractmethod
    def _set_sensor(self, value: Device) -> None:
        """Set the temperature sensor for the zone (or for the DHW)."""
        raise NotImplementedError

    @property
    def name(self) -> Optional[str]:
        """Return the name of the zone/DHW."""
        return self._name

    @property
    @abstractmethod
    def mode(self) -> dict:
        """Return the operating mode of the zone/DHW ({mode, setpoint, until})."""
        raise NotImplementedError

    @property
    @abstractmethod
    def setpoint(self) -> Optional[float]:
        """Return the target temperature of the zone/DHW."""
        raise NotImplementedError

    # @setpoint.setter
    # @abstractmethod
    # def setpoint(self, value: float) -> None:  # TODO: also mode, name attrs
    #     """Set the target temperature of the zone/DHW."""
    #     raise NotImplementedError

    @property
    @abstractmethod
    def temperature(self) -> Optional[float]:
        """Return the measured temperature of the zone/DHW.

        The controller, not the temperature sensor, is the source of this data.
        """
        raise NotImplementedError

    # @property
    # @abstractmethod
    # def type(self) -> str:
    #     """Return the type of the zone/DHW (e.g. electric_zone, stored_dhw)."""
    #     raise NotImplementedError

    # @abstractmethod
    # def _set_type(self, value: str) -> None:
    #     """Set the type of the zone/DHW (e.g. electric_zone, stored_dhw)."""
    #     raise NotImplementedError


class DhwZone(ZoneBase, HeatDemand):
    """The DHW class.

    FC - 0008, 0009, 1100, 3150, 3B00, (& rare: 0001, 1FC9)
    """

    def __init__(self, controller, sensor=None, relay=None) -> None:
        super().__init__(controller, "FA")

        controller.dhw = self

        self._sensor = None
        self._dhw_valve = None
        self._htg_valve = None
        self.heating_type = "DHW"
        self._name = "Stored DHW"

        self._dhw_mode = {}
        self._dhw_params = {}
        self._temperature = None
        self._relay_demand = None
        self._relay_failsafe = None

        self._discover()  # should be last thing in __init__()

    def _discover(self, discover_flags=DISCOVER_ALL) -> None:
        if self._gwy.config["disable_discovery"]:
            return
        # super()._discover()

        # if False and __dev_mode__ and self.idx == "FA":  # dev/test code
        #     self.async_set_override(state="On")

        if discover_flags & DISCOVER_SCHEMA:
            [  # 000C: find the DHW relay(s), if any, see: CODE_000C_DEVICE_TYPE
                self._send_cmd("000C", payload=dev_type)
                for dev_type in ("000D", "000E", "010E")  # for DHW sensor, relay(s)
            ]

        if discover_flags & DISCOVER_PARAMS:
            for code in ("10A0",):
                self._send_cmd(code, payload="00")  # payload="00" or "0000", not "FA"

        if discover_flags & DISCOVER_STATUS:
            for code in ("1260", "1F41"):
                self._send_cmd(code, payload="00")  # payload="00" or "0000", not "FA"

    def _proc_msg(self, msg) -> None:
        super()._proc_msg(msg)

        if msg.code == "0008":
            self._relay_demand = msg.payload["relay_demand"]
        elif msg.code == "0009":
            self._relay_failsafe = msg.payload
        elif msg.code == "10A0":
            self._dhw_params = {
                x: msg.payload[x] for x in ("setpoint", "overrun", "differential")
            }
        elif msg.code == "1260":
            self._temperature = msg.payload["temperature"]
        elif msg.code == "1F41":
            self._setpoint_status = {
                x: msg.payload[x] for x in ("active", "dhw_mode", "until")
            }
        else:
            assert False, "Unknown packet code"

    @property
    def schema(self) -> dict:
        """Return the stored HW's schema."""

        return {
            ATTR_DHW_SENSOR: self._sensor.id if self._sensor else None,
            ATTR_DHW_VALVE: self._dhw_valve.id if self._dhw_valve else None,
            ATTR_DHW_VALVE_HTG: self._htg_valve.id if self._htg_valve else None,
        }

    @property  # setpoint, config, mode (not schedule)
    def params(self) -> dict:
        """Return the stored HW's configuration (excl. schedule)."""

        return {
            "dhw_params": self._dhw_params,
        }

    @property  # temp, open_windows
    def status(self) -> dict:
        """Return the stored HW's current state."""

        return {
            "temperature": self._temperature,
            "dhw_mode": self._dhw_mode,
        }

    @property
    def sensor(self) -> Device:
        """Blah it now.

        Check and Verb the DHW sensor (07:) of this system/CTL (if there is one).

        There is only 1 way to find a controller's DHW sensor:
        1.  The 10A0 RQ/RP *from/to a 07:* (1x/4h)

        The RQ is initiated by the DHW, so is not authorative (the CTL will RP any RQ).
        The I/1260 is not to/from a controller, so is not useful.
        """  # noqa: D402

        # 07:38:39.124 047 RQ --- 07:030741 01:102458 --:------ 10A0 006 00181F0003E4
        # 07:38:39.140 062 RP --- 01:102458 07:030741 --:------ 10A0 006 0018380003E8

        if "10A0" in self._msgs:
            return self._msgs["10A0"].dst.addr

        return self._sensor

    def _set_sensor(self, device: Device) -> None:
        """Set the sensor for this DHW (must be: 07:)."""

        if not isinstance(device, Device) or device.type != "07":
            raise TypeError(f"Invalid device type for DHW sensor: {device}")

        if self._sensor is None:
            self._sensor = device
            device._set_domain(dhw=self)  # TODO: check have same controller

        elif self._sensor != device:
            raise CorruptStateError(f"DHW sensor changed: {self._sensor} to {device}")

    @property
    def hotwater_valve(self) -> Device:
        return self._dhw_valve

    @hotwater_valve.setter
    def hotwater_valve(self, device: Device) -> None:
        if not isinstance(device, Device) or device.type != "13":
            raise TypeError

        if self._dhw_valve is not None and self._dhw_valve != device:
            raise CorruptStateError("The DHW HW valve has changed")
        # elif device.evo is not None and device.evo != self:
        #     raise LookupError  #  do this in add_devices

        if self._dhw_valve is None:
            self._dhw_valve = device
            device._set_domain(dhw=self)

    @property
    def heating_valve(self) -> Device:
        return self._htg_valve

    @heating_valve.setter
    def heating_valve(self, device: Device) -> None:
        if not isinstance(device, Device) or device.type != "13":
            raise TypeError

        if self._htg_valve is not None and self._htg_valve != device:
            raise CorruptStateError("The DHW heating valve has changed")
        # elif device.evo is not None and device.evo != self:
        #     raise LookupError  #  do this in add_devices

        if self._htg_valve is None:
            self._htg_valve = device
            device._set_domain(dhw=self)

    @property
    def relay_demand(self) -> Optional[float]:  # 0008
        return self._relay_demand

    @property  # only seen with FC, but seems should pair with 0008?
    def relay_failsafe(self) -> Optional[float]:  # 0009
        return self._relay_failsafe

    @property
    def dhw_config(self) -> dict:  # 10A0
        return self._dhw_params

    @property
    def mode(self) -> dict:  # 1F41
        return self._dhw_mode

    @property
    def setpoint(self) -> Optional[float]:  # 1F41
        return self._dhw_mode.get("setpoint")

    @property
    def temperature(self) -> Optional[float]:  # 1260
        return self._temperature

    async def cancel_override(self) -> bool:  # 1F41
        """Reset the DHW to follow its schedule."""
        return False

    async def set_override(self, mode=None, state=None, until=None) -> bool:
        """Force the DHW on/off for a duration, or indefinitely.

        Use until = ? for 1hr boost (obligates on)
        Use until = ? for until next scheduled on/off
        Use until = None for indefinitely
        """
        # 053  I --- 01:145038 --:------ 01:145038 1F41 012 00 01 04 FFFFFF 1E061B0607E4
        # 048  I --- 01:145038 --:------ 01:145038 1F41 012 00 00 04 FFFFFF 1E061B0607E4

        # if mode is None and until is None:
        #     mode = "00" if setpoint is None else "02"  # Follow, Permanent
        # elif mode is None:  # and until is not None
        #     mode = "04"  # Temporary
        # elif isinstance(mode, int):
        #     mode = f"{mode:02X}"
        # elif not isinstance(mode, str):
        #     raise TypeError("Invalid zone mode")
        # elif mode in ZONE_MODE_LOOKUP:
        #     mode = ZONE_MODE_LOOKUP[mode]

        # if mode not in ZONE_MODE_MAP:
        #     raise ValueError("Unknown zone mode")

        # if state is None and until is None:
        #     state = "01"
        # elif state is None:  # and until is not None
        #     state = "01"
        # elif isinstance(state, int):
        #     mode = f"{mode:02X}"
        # elif isinstance(state, bool):
        #     mode = "01" if mode is True else "00"
        # elif not isinstance(mode, str):
        #     raise TypeError("Invalid DHW state")
        # elif state in DHW_STATE_LOOKUP:
        #     state = DHW_STATE_LOOKUP[mode]

        if state not in DHW_STATE_MAP:
            raise ValueError("Unknown DHW state")

        if until is None:
            payload = f"00{state}{mode}FFFFFF"
        else:  # required only by: 04, Temporary, ignored by others
            payload = f"00{state}{mode}FFFFFF{_dtm(until)}"

        self._send_cmd("1F41", verb=" W", payload=payload)
        return False

    async def reset_config(self) -> bool:  # 10A0
        """Reset the DHW parameters to their default values."""
        return False

    async def set_config(self, setpoint, overrun=None, differential=None) -> bool:
        """Set the DHW parameters."""
        return False


class Zone(ZoneBase):
    """The Zone class."""

    def __init__(self, controller, zone_idx, sensor=None, actuators=None) -> None:
        """Create a zone.

        The type of zone may not be known at instantiation. Even when it is known, zones
        are still created without a type before they are subsequently promoted, so that
        both schemes (e.g. eavesdropping, vs probing) are the same.

        In addition, an electric zone may subsequently turn out to be a zone valve zone.
        """
        super().__init__(controller, zone_idx)

        assert (
            zone_idx not in controller.zone_by_idx
        ), "Duplicate zone idx on controller"
        if int(zone_idx, 16) >= self._gwy.config["max_zones"]:
            raise ValueError  # TODO: better to aloow to disable via assert?

        controller.zones.append(self)
        controller.zone_by_idx[zone_idx] = self
        # controller.zone_by_name[self.name] = self

        self.devices = []
        self.device_by_id = {}
        self._sensor = None

        # attributes for .params and .status
        self._mode = None
        self._setpoint = None
        self._temperature = None
        self._window_open = None
        self._zone_config = None

        self._schedule = Schedule(self)

        self._discover()

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        if self._gwy.config["disable_discovery"]:
            return
        # super()._discover()

        if __dev_mode__ and self.idx == "99":  # dev/test code
            asyncio.create_task(  # TODO: test/dev only
                self.async_cancel_override()
                # self.async_set_override(
                #     setpoint=15.9,
                #     mode="AdvancedOverride",
                #     # until=dt_now() + timedelta(minutes=120)
                # )
            )

        # TODO: add code to determine zone type if it doesn't have one, using 0005s

        [  # 000C: find the sensor and the actuators, if any
            self._send_cmd("000C", payload=f"{self.idx}{dev_type}")
            for dev_type in ("00", "04")  # CODE_0005_ZONE_TYPE
            # for dev_type, description in CODE_000C_DEVICE_TYPE.items()
            # if description is not None
        ]

        # start collecting the schedule
        # self._schedule.req_schedule()  # , restart=True) start collecting schedule

        for code in ("0004",):
            self._send_cmd(code, payload=f"{self.idx}00")

        for code in ("000A", "2349", "30C9"):  # sadly, no 3150
            self._send_cmd(code, payload=self.idx)

    def _proc_msg(self, msg) -> None:
        super()._proc_msg(msg)

        if msg.code == "0004":
            self._name = msg.payload.get("name")

        # not UFH (it seems), but ELE or VAL; and possibly a MIX support 0008 too
        elif msg.code in ("0008", "0009"):  # TODO: how to determine is/isn't MIX?
            assert msg.src.type in ("01", "13")  # 01 as a stat
            assert self._zone_type in (None, "ELE", "VAL")

            if self._zone_type is None:
                self._set_type("ELE")  # might eventually be: "VAL"

        elif msg.code == "000A":
            payload = msg.payload if msg.is_array else [msg.payload]
            self._zone_config = {
                k: v
                for z in payload
                for k, v in z.items()
                if z["zone_idx"] == self.idx and k[:1] != "_" and k != "zone_idx"
            }

        elif msg.code == "0404" and msg.verb == "RP":
            _LOGGER.error("Zone(%s).update: Received RP/0404 (schedule)", self.id)
            self._schedule.add_fragment(msg)
            self._schedule.req_fragment()  # do only if we self._schedule.req_schedule()

        elif msg.code == "2309":
            payload = msg.payload if msg.is_array else [msg.payload]
            self._setpoint = {
                k: v for z in payload for k, v in z.items() if z["zone_idx"] == self.idx
            }["setpoint"]

        elif msg.code == "2349":
            self._mode = {
                k: v
                for k, v in msg.payload.items()
                if k in ("mode", "setpoint", "until")
            }
            self._setpoint = msg.payload["setpoint"]

        elif msg.code == "30C9":  # required for sensor matching
            assert msg.src.type in DEVICE_HAS_ZONE_SENSOR + ("01",)
            payload = msg.payload if msg.is_array else [msg.payload]
            self._temperature = {
                k: v for z in payload for k, v in z.items() if z["zone_idx"] == self.idx
            }["temperature"]

        elif msg.code == "3150":  # TODO: and msg.verb in (" I", "RP")?
            assert msg.src.type in ("02", "04", "13")
            assert self._zone_type in (None, "RAD", "UFH", "VAL")  # ELE don't have 3150

            if msg.src.type in ("02", "04", "13"):
                zone_type = ZONE_CLASS_MAP[msg.src.type]
                self._set_type("VAL" if zone_type == "ELE" else zone_type)

        # elif "zone_idx" in msg.payload:
        #     pass

        # elif msg.code not in ("FFFF"):
        #     assert False, "Unknown packet code"

    @property  # id, type
    def schema(self) -> dict:
        """Return the zone's schema."""

        return {
            ATTR_ZONE_TYPE: self.heating_type,
            ATTR_ZONE_SENSOR: self._sensor.id if self._sensor else None,
            ATTR_DEVICES: [d.id for d in self.devices],
        }

    @property  # setpoint, config, mode (not schedule)
    def params(self) -> dict:
        """Return the zone's configuration (excl. schedule)."""

        ATTR_NAME = "name"
        ATTR_MODE = "mode"
        ATTR_CONFIG = "zone_config"

        return {
            ATTR_NAME: self.name,
            ATTR_MODE: self.mode,
            ATTR_CONFIG: self.zone_config,
        }

    @property
    def status(self) -> dict:
        """Return the zone's current state."""

        return {
            ATTR_SETPOINT: self.setpoint,
            ATTR_TEMP: self.temperature,
        }

    @property
    def sensor(self) -> Device:
        return self._sensor

    def _set_sensor(self, device: Device):
        """Set the sensor for this zone (one of: 01:, 03:, 04:, 12:, 22:, 34:)."""

        sensor_types = ("01", "03", "04", "12", "22", "34")
        if not isinstance(device, Device) or device.type not in sensor_types:
            raise TypeError(f"Invalid device type for zone sensor: {device}")

        if self._sensor is None:
            self._sensor = device  # if TRV, zone type likely (but not req'd) RAD
            device._set_domain(zone=self)  # TODO: check have same controller

        elif self._sensor is not device:
            raise CorruptStateError(f"zone sensor changed: {self._sensor} to {device}")

    @property
    def heating_type(self) -> Optional[str]:
        """TODO.

        There are three ways to determine the type of a zone:
        1. Use a 0005 packet (deterministic)
        2. Eavesdrop (non-deterministic, slow to converge)
        3. via a config file (a schema)
        """

        if self._zone_type is not None:  # isinstance(self, ???)
            return ZONE_TYPE_MAP.get(self._zone_type)

        # TODO: actuators
        dev_types = [d.type for d in self.devices if d.type in ("02", "04", "13")]

        if "02" in dev_types:
            zone_type = "UFH"
        elif "13" in dev_types and "3150" in self._msgs:
            zone_type = "VAL"
        elif "13" in dev_types:
            zone_type = "ELE"  # could still be a VAL
        # elif "??" in dev_types:  # TODO:
        #     zone_type = "MIX"
        elif "04" in dev_types:  # beware edge case: TRV as sensor for a non-RAD zone
            zone_type = "RAD"
        else:
            zone_type = None

        if zone_type is not None:
            self._set_type(zone_type)

        return ZONE_TYPE_MAP.get(self._zone_type)

    def _set_type(self, zone_type: str):
        """Set the zone's type, after validating it.

        There are two possible sources for the type of a zone:
        1. eavesdropping packet codes
        2. analyzing child devices

        Both will execute a zone.type = type (i.e. via this setter).
        """

        _type = ZONE_TYPE_SLUGS.get(zone_type, zone_type)
        if _type not in ZONE_CLASSES:
            raise ValueError(f"Not a known zone type: {zone_type}")

        if self._zone_type is not None:
            if self._zone_type != _type and (
                self._zone_type != "ELE" and _type != "VAL"
            ):
                raise CorruptStateError(
                    f"Zone {self} has a mismatched type: "
                    f"old={self._zone_type}, new={_type}",
                )

        self._zone_type = _type
        self.__class__ = ZONE_CLASSES[_type]
        _LOGGER.debug("Zone %s: type now set to %s", self.id, self._zone_type)

    @property
    def zone_config(self) -> Optional[dict]:  # 000A
        return self._zone_config

    @property
    def mode(self) -> Optional[dict]:  # 2349
        # result = self._get_msg_value("2349")
        # self._mode = (
        #     {k: v for k, v in result.items() if k != "zone_idx"} if result else None
        # )

        return self._mode

    @property
    def setpoint(self) -> Optional[float]:  # 2309 (2349 is a superset of 2309)
        msg = self._ctl._msgs.get("2309")
        if msg is not None:
            self._setpoint = {
                k: v
                for z in msg.payload
                for k, v in z.items()
                if z["zone_idx"] == self.idx
            }.get("setpoint")
        return self._setpoint

    @property
    def temperature(self) -> Optional[float]:  # 30C9
        # TODO: this wont wokr if the controller is the sensor
        # if self.sensor and self.sensor.temperature:
        #     return self.sensor.temperature

        msg = self._ctl._msgs.get("30C9")
        if msg is not None:
            self._temperature = {
                k: v
                for z in msg.payload
                for k, v in z.items()
                if z["zone_idx"] == self.idx
            }.get("temperature")
        return self._temperature

    def schedule(self, force_update=False) -> Optional[dict]:
        """Return the schedule if any."""
        return self._schedule.schedule if self._schedule else None

    async def cancel_override(self) -> None:  # 2349
        """Revert to following the schedule."""
        await self.set_override()

    async def frost_protect(self) -> None:  # 2349
        """Set the zone to the lowest possible setpoint, indefinitely."""
        await self.set_override(mode="02", setpoint=5)  # TODO

    async def set_override(self, mode=None, setpoint=None, until=None) -> None:
        """Override the setpoint for a specified duration, or indefinitely.

        The setpoint has a resolution of 0.1 C. If a setpoint temperature is required,
        but none is provided, the controller will use the maximum possible value.

        The until has a resolution of 1 min.

        Incompatible combinations:
          - mode == Follow & setpoint not None (will silently ignore setpoint)
          - mode == Temporary & until is None (will silently drop W packet)
        """

        if mode is None and until is None:
            mode = "00" if setpoint is None else "02"  # Follow, Permanent
        elif mode is None:  # and until is not None
            mode = "04"  # Temporary
        elif isinstance(mode, int):
            mode = f"{mode:02X}"
        elif not isinstance(mode, str):
            raise TypeError("Invalid zone mode")
        elif mode in ZONE_MODE_LOOKUP:
            mode = ZONE_MODE_LOOKUP[mode]

        if mode not in ZONE_MODE_MAP:
            raise ValueError("Unknown zone mode")

        setpoint = _temp(setpoint)  # None means max, if a temp is required

        if until is None:
            mode = "01" if mode == "04" else mode
            payload = f"{self.idx}{setpoint}{mode}FFFFFF"
        else:  # required only by: 04, Temporary, ignored by others
            payload = f"{self.idx}{setpoint}{mode}FFFFFF{_dtm(until)}"

        self._send_cmd("2349", verb=" W", payload=payload)


class ZoneHeatDemand:  # not all zone types call for heat
    """Not all zones call for heat."""

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        if self._gwy.config["disable_discovery"]:
            return
        super()._discover()

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("12B0")  # , payload=self.idx

    @property
    def heat_demand(self) -> Optional[float]:  # 3150
        demands = [
            d.heat_demand
            for d in self.devices
            if hasattr(d, "heat_demand") and d.heat_demand is not None
        ]
        return max(demands + [0]) if demands else None

    @property
    def status(self) -> dict:
        return {**super().status, ATTR_HEAT_DEMAND: self.heat_demand}


class EleZone(Zone):  # Electric zones (do *not* call for heat)
    """Base for Electric Heat zones.

    For a small (5A) electric load controlled by a BDR91 (never calls for heat).
    """

    def _proc_msg(self, msg) -> None:
        super()._proc_msg(msg)

        # ZV zones are Elec zones that also call for heat; ? and also 1100/unkown_0 = 00
        if msg.code == "3150":
            self._set_type("VAL")

        # if msg.code == "FFFF":
        #     pass
        # else:
        #     assert False, "Unknown packet code"

    @property
    def actuator_enabled(self) -> Optional[bool]:  # 3EF0
        return self._get_msg_value("3EF0")  # , "actuator_enabled"

    @property
    def actuator_state(self) -> Optional[float]:  # 3EF1
        return self._get_msg_value("3EF1")

    @property
    def status(self) -> dict:
        return {
            **super().status,
            "actuator_enabled": self.actuator_enabled,
            "actuator_state": self.actuator_state,
        }


class ValZone(ZoneHeatDemand, EleZone):  # Zone valve zones
    """Base for Zone Valve zones.

    For a motorised valve controlled by a BDR91 (will also call for heat).
    """


class RadZone(ZoneHeatDemand, Zone):  # Radiator zones
    """Base for Radiator Valve zones.

    For radiators controlled by HR92s or HR80s (will also call for heat).
    """

    def _proc_msg(self, msg) -> None:
        super()._proc_msg(msg)

        if msg.code == "12B0":
            self._window_open = msg.payload["window_open"]

        # if msg.code == "FFFF":
        #     pass
        # else:
        #     assert False, "Unknown packet code"

    # 3150 (heat_demand) but no 0008 (relay_demand)

    @property
    def window_open(self) -> Optional[bool]:  # 12B0
        return self._window_open

    @property
    def status(self) -> dict:
        return {
            **super().status,
            ATTR_OPEN_WINDOW: self.window_open,
        }


class UfhZone(ZoneHeatDemand, Zone):  # UFH zones
    """Base for Underfloor Heating zones.

    For underfloor heating controlled by an HCE80 or HCC80 (will also call for heat).
    """

    @property
    def ufh_setpoint(self) -> Optional[float]:  # 3B00
        return self._get_msg_value("22C9")

    @property
    def status(self) -> dict:
        return {**super().status, "ufh_setpoint": self.ufh_setpoint}


class MixZone(ZoneHeatDemand, Zone):  # Mix valve zones
    """Base for Mixing Valve zones.

    For a modulating valve controlled by a HM80 (will also call for heat).
    """

    @property
    def mix_config(self) -> dict:
        attrs = ("pump_run_time", "actuator_run_time", "min_flow_temp", "max_flow_temp")
        return {x: self._get_msg_value("1030", x) for x in attrs}

    @property
    def params(self) -> dict:
        return {**super().status, "mix_config": self.mix_config}


ZONE_CLASSES = {
    "RAD": RadZone,
    "ELE": EleZone,
    "VAL": ValZone,
    "UFH": UfhZone,
    "MIX": MixZone,
    "DHW": DhwZone,
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Logging utility."""

import ctypes
from datetime import datetime as dt
import logging
import shutil
import sys
import time

# from logging.handlers import TimedRotatingFileHandler

from .const import __dev_mode__

BASIC_FMT = "%(asctime)s.%(msecs)03d %(message)s"
BASIC_DATEFMT = "%H:%M:%S"
BASIC_LEVEL = logging.INFO

BANDW_SUFFIX = "%(error_text)s%(comment)s"
COLOR_SUFFIX = "%(red)s%(error_text)s%(cyan)s%(comment)s"

try:
    from colorlog import ColoredFormatter  # default_log_colors
except ModuleNotFoundError:
    COLOR_SUFFIX = BANDW_SUFFIX

# basicConfig must be called after importing colorlog to ensure its handlers wrap the
# correct streams
logging.basicConfig(level=BASIC_LEVEL, format=BASIC_FMT, datefmt=BASIC_DATEFMT)

# HH:MM:SS.sss vs YYYY-MM-DDTHH:MM:SS.ssssss, shorter format for the console
CONSOLE_COLS = int(shutil.get_terminal_size(fallback=(2e3, 24)).columns - 1)
CONSOLE_FMT = "%(time).12s " + f"%(message).{CONSOLE_COLS - 13}s"
PKT_LOG_FMT = "%(date)sT%(time)s %(_packet)s"
MSG_LOG_FMT = "%(date)sT%(time)s %(message)s"

if __dev_mode__:
    CONSOLE_FMT = MSG_LOG_FMT  # Do this to have longer-format console messages
# How to strip ASCII colour from a text file:
#   sed -r "s/\x1B\[(([0-9]{1,2})?(;)?([0-9]{1,2})?)?[m,K,H,f,J]//g" file_name

LOG_COLOURS = {
    "DEBUG": "white",
    "INFO": "green",
    "WARNING": "yellow",
    "ERROR": "bold_red",
    "CRITICAL": "bold_red",
}  # default_log_colors


class FILETIME(ctypes.Structure):
    """Data structure for GetSystemTimePreciseAsFileTime()."""

    _fields_ = [("dwLowDateTime", ctypes.c_uint), ("dwHighDateTime", ctypes.c_uint)]


def dt_now() -> dt:
    """Return the time now as a UTC datetime object."""
    return dt.fromtimestamp(time_time())


def dt_str() -> str:
    """Return the time now as a isoformat string."""
    now = time_time()
    mil = f"{now%1:.6f}".lstrip("0")
    return time.strftime(f"%Y-%m-%dT%H:%M:%S{mil}", time.localtime(now))


def time_time() -> float:
    """Return the number of seconds since the Unix epoch.

    Return an accurate value, even for Windows-based systems.
    """  # see: https://www.python.org/dev/peps/pep-0564/
    if sys.platform != "win32":
        return time.time()  # since 1970-01-01T00:00:00Z, time.gmtime(0)
    file_time = FILETIME()
    ctypes.windll.kernel32.GetSystemTimePreciseAsFileTime(ctypes.byref(file_time))
    _time = (file_time.dwLowDateTime + (file_time.dwHighDateTime << 32)) / 1e7
    return _time - 134774 * 24 * 60 * 60  # otherwise, is since 1601-01-01T00:00:00Z


def set_logging(
    logger,
    stream=sys.stderr,
    cons_fmt=CONSOLE_FMT,
    file_fmt=MSG_LOG_FMT,
    file_name=None,
) -> None:
    """Create/configure handlers, formatters, etc."""
    logger.propagate = False

    if COLOR_SUFFIX == BANDW_SUFFIX:
        formatter = logging.Formatter(fmt=cons_fmt)
    else:
        formatter = ColoredFormatter(
            f"%(log_color)s{cons_fmt}", reset=True, log_colors=LOG_COLOURS
        )

    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(formatter)
    handler.setLevel(logging.WARNING)
    # handler.addFilter(DebugFilter())

    logger.addHandler(handler)

    if stream == sys.stdout:
        handler = logging.StreamHandler(stream=stream)
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)  # TODO: should be WARNING, but breaks logging
        handler.addFilter(InfoFilter())

        logger.addHandler(handler)

    if file_name:
        # if log_rotate_days:
        #     err_handler = logging.handlers.TimedRotatingFileHandler(
        #         err_log_file_name, when="midnight", backupCount=log_rotate_days
        #     )
        # else:
        #     err_handler = logging.FileHandler(err_log_path, mode="w", delay=True)

        # err_handler.setLevel(logging.INFO if verbose else logging.WARNING)
        # err_handler.setFormatter(logging.Formatter(fmt, datefmt=datefmt))

        handler = logging.FileHandler(file_name)
        handler.setFormatter(logging.Formatter(fmt=file_fmt))
        handler.setLevel(logging.DEBUG)
        handler.addFilter(DebugFilter())  # TODO: was InfoFilter()

        logger.addHandler(handler)


class InfoFilter(logging.Filter):
    """Log only INFO-level messages."""

    def filter(self, record) -> bool:
        """Filter out all but INFO/DEBUG packets."""
        return record.levelno in (logging.INFO, logging.DEBUG)


class DebugFilter(logging.Filter):
    """Don't Log DEBUG-level messages."""

    def filter(self, record) -> bool:
        """Filter out all DEBUG packets."""
        return record.levelno != logging.DEBUG  # TODO: use less than / more than?



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Evohome serial discovery scripts."""

import asyncio
import logging

from .command import Command
from .const import __dev_mode__  # , CODE_SCHEMA


_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


async def periodic(cmd_que, cmd, count=1440, interval=5):
    if count <= 0:
        while True:
            await asyncio.sleep(interval)
            cmd_que.put_nowait(cmd)

    else:
        for _ in range(count):
            await asyncio.sleep(interval)
            cmd_que.put_nowait(cmd)


def poll_device(cmd_que, device_id):
    for code in ("0008", "3EF1"):
        cmd = Command("RQ", device_id, code, "00", retry_limit=0)
        _ = asyncio.create_task(periodic(cmd_que, cmd, count=0, interval=15))


def probe_device(cmd_que, device_id):
    for code in ("0016", "1FC9"):  # payload 0000 OK for both these
        cmd = Command("RQ", device_id, code, "0000", retry_limit=9)
        _ = asyncio.create_task(periodic(cmd_que, cmd, count=1))

    if device_id.startswith("13"):
        for code in ("0008", "1100", "3EF1"):
            cmd = Command("RQ", device_id, code, "00", retry_limit=9)
            _ = asyncio.create_task(periodic(cmd_que, cmd, count=1))

    # for code in sorted(CODE_SCHEMA):
    #     cmd = Command("RQ", device_id, code, "0000", retry_limit=5)
    #     _ = asyncio.create_task(periodic(cmd_que, cmd, count=1))


# if self.config.get("evofw_flag") and "evofw3" in raw_pkt.packet:
#     # !V, !T - print the version, or the current mask
#     # !T00   - turn off all mask bits
#     # !T01   - cause raw data for all messages to be printed
#     await manager.put_pkt(self.config["evofw_flag"], _LOGGER)


# # used for development only...
# for code in range(0x4000):
#     # cmd = Command("RQ", "01:145038", f"{code:04X}", "0000")
#     cmd = Command("RQ", "13:035462", f"{code:04X}", "0000")
#     await destination.put_pkt(cmd, _LOGGER)
#     if code % 0x10 == 0:
#         await asyncio.sleep(15)  # 10 too short - 15 seconds works OK


# # used for development only...
# for payload in ("0000", "0100", "F8", "F9", "FA", "FB", "FC", "FF"):
#     cmd = Command("RQ", "01:145038", "11F0", payload)
#     await destination.put_pkt(cmd, _LOGGER)
#     cmd = Command("RQ", "13:035462", "11F0", payload)
#     await destination.put_pkt(cmd, _LOGGER)


# for device_type in ("0D", "0E", "0F"):  # CODE_000C_DEVICE_TYPE:
#     cmd = Command("RQ", "01:145038", "000C", f"00{device_type}")
#     await manager.put_pkt(cmd, _LOGGER)


# for z in range(4):
#     for x in range(12):
#         cmd = Command("RQ", "01:145038", "000C", f"{z:02X}{x:02X}")
#         await manager.put_pkt(cmd, _LOGGER)


# for p in ("00", "01", "FF", "0000", "0100", "FF00"):
#     for c in ("0003", "0007", "000B", "000D", "000F"):
#         cmd = Command("RQ", "01:145038", c, f"0008{p}")
#         print(cmd)
#         await manager.put_pkt(cmd, _LOGGER)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""Message processor."""

from datetime import datetime as dt
import logging
import re
from typing import Any, Optional, Union

from . import parsers
from .const import (
    ATTR_HTG_CONTROL,
    ATTR_DHW_SENSOR,
    ATTR_DHW_VALVE_HTG,
    ATTR_DHW_VALVE,
    ATTR_ZONE_SENSOR,
    CODE_MAP,
    DEVICE_TYPES,
    MSG_FORMAT_10,
    MSG_FORMAT_18,
    NON_DEVICE,
    NUL_DEVICE,
    CODE_0005_ZONE_TYPE,
    ZONE_TYPE_SLUGS,
    Address,
    __dev_mode__,
)
from .devices import Device
from .exceptions import CorruptPayloadError

_LOGGER = logging.getLogger(__name__)
if __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)


def exception_handler(func):
    """xxx."""

    def wrapper(*args, **kwargs) -> Optional[Any]:
        """xxx."""
        try:
            return func(*args, **kwargs)
        except (AssertionError, NotImplementedError) as err:
            msg = args[0]
            _LOGGER.exception(
                "%s < %s", msg._pkt, err.__class__.__name__, extra=msg.__dict__
            )
            raise
        # TODO: this shouldn't be required?
        except (AttributeError, LookupError, TypeError, ValueError) as err:
            msg = args[0]
            _LOGGER.error(
                "%s < %s", msg._pkt, err.__class__.__name__, extra=msg.__dict__
            )
            raise

    return wrapper


class Message:
    """The message class."""

    def __init__(self, gateway, pkt) -> None:
        """Create a message, assumes a valid packet."""
        self._gwy = gateway
        self._pkt = packet = pkt.packet

        # prefer Device(s) but Address(es) will do
        self.src = gateway.device_by_id.get(pkt.src_addr.id, pkt.src_addr)
        self.dst = gateway.device_by_id.get(pkt.dst_addr.id, pkt.dst_addr)

        self.devs = pkt.addrs
        self.date = pkt.date
        self.time = pkt.time
        self.dtm = dt.fromisoformat(f"{pkt.date}T{pkt.time}")

        self.rssi = packet[0:3]
        self.verb = packet[4:6]
        self.seqn = packet[7:10]  # sequence number (as used by 31D9)?
        self.code = packet[41:45]

        self.len = int(packet[46:49])  # TODO:  is useful? / is user used?
        self.raw_payload = packet[50:]

        self._payload = self._str = None

        self._is_valid = self._is_array = self._is_fragment = None
        self._is_valid = self.is_valid

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return self._pkt

    def __str__(self) -> str:
        """Return a brief readable string representation of this object."""

        def display_name(dev: Union[Address, Device]) -> str:
            """Return a formatted device name, uses a friendly name if there is one."""
            if dev is NON_DEVICE:
                return f"{'':<10}"

            if dev is NUL_DEVICE:
                return "NUL:------"

            if dev.id in self._gwy.known_devices:
                if self._gwy.known_devices[dev.id].get("friendly_name"):
                    return self._gwy.known_devices[dev.id]["friendly_name"]

            return f"{DEVICE_TYPES.get(dev.type, f'{dev.type:>3}')}:{dev.id[3:]}"

        if self._str is not None:
            return self._str

        if not self.is_valid:
            return

        _format = MSG_FORMAT_18 if self._gwy.config["known_devices"] else MSG_FORMAT_10

        if self.src.id == self.devs[0].id:
            src = display_name(self.src)
            dst = display_name(self.dst) if self.dst is not self.src else ""
        else:
            src = ""
            dst = display_name(self.src)

        code = CODE_MAP.get(self.code, f"unknown_{self.code}")
        payload = self.raw_payload if self.len < 4 else f"{self.raw_payload[:5]}..."[:9]

        self._str = _format.format(src, dst, self.verb, code, payload, self._payload)
        return self._str

    def __eq__(self, other) -> bool:
        return all(
            self.verb == other.verb,
            self.code == other.code,
            self.src.id == other.src.id,
            self.dst.id == other.dst.id,
            self.raw_payload == other.raw_payload,
        )

    @property
    def payload(self) -> Any:  # Any[dict, List[dict]]:
        """Return the payload."""
        if self._payload is not None:
            return self._payload

        if self.is_valid:
            if self.code != "000C":  # TODO: assert here, or in is_valid()
                assert self.is_array == isinstance(self._payload, list)
            return self._payload

    @property
    def is_array(self) -> bool:
        """Return True if the message's raw payload is an array.

        Note that the corresponding parsed payload may not match, e.g. the 000C payload
        is not a list.

        Does not require a valid payload.
        """

        if self._is_array is not None:
            return self._is_array

        if self.code in ("000C", "1FC9"):  # also: 0005?
            # grep -E ' (I|RP).* 000C '  #  from 01:/30: (VMS) only
            # grep -E ' (I|RP).* 1FC9 '  #  from 01:/13:/other (not W)
            self._is_array = self.verb in (" I", "RP")

        elif self.verb not in (" I", "RP") or self.src.id != self.dst.id:
            self._is_array = False

        # 045  I --- 01:158182 --:------ 01:158182 0009 003 0B00FF (or: FC00FF)
        # 045  I --- 01:145038 --:------ 01:145038 0009 006 FC00FFF900FF
        elif self.code in ("0009") and self.src.type == "01":
            # grep -E ' I.* 01:.* 01:.* 0009 [0-9]{3} F' (and: grep -v ' 003 ')
            self._is_array = self.verb == " I" and self.raw_payload[:1] == "F"

        elif self.code in ("000A", "2309", "30C9") and self.src.type == "01":
            # grep ' I.* 01:.* 01:.* 000A '
            # grep ' I.* 01:.* 01:.* 2309 ' | grep -v ' 003 '  # TODO: some non-arrays
            # grep ' I.* 01:.* 01:.* 30C9 '
            self._is_array = self.verb == " I" and self.src.id == self.dst.id

        # 055  I --- 02:001107 --:------ 02:001107 22C9 024 0008340A28010108340A...
        # 055  I --- 02:001107 --:------ 02:001107 22C9 006 0408340A2801
        # 055  I --- 02:001107 --:------ 02:001107 3150 010 00640164026403580458
        # 055  I --- 02:001107 --:------ 02:001107 3150 010 00000100020003000400
        elif self.code in ("22C9", "3150") and self.src.type == "02":
            # grep -E ' I.* 02:.* 02:.* 22C9 '
            # grep -E ' I.* 02:.* 02:.* 3150' | grep -v FC
            self._is_array = self.verb == " I" and self.src.id == self.dst.id
            self._is_array = self._is_array if self.raw_payload[:1] != "F" else False

        # 095  I --- 23:100224 --:------ 23:100224 2249 007 007EFF7EFFFFFF
        # 095  I --- 23:100224 --:------ 23:100224 2249 007 007EFF7EFFFFFF
        elif self.code in ("2249") and self.src.type == "23":
            self._is_array = self.verb == " I" and self.src.id == self.dst.id
            # self._is_array = self._is_array if self.raw_payload[:1] != "F" else False

        else:
            self._is_array = False

        return self._is_array

    @property
    def is_fragment_WIP(self) -> bool:
        """Return True if the raw payload is a fragment of a message."""

        # if not self._is_valid:
        #     return
        if self._is_fragment is not None:
            return self._is_fragment

        # packets have a maximum length of 48 (decimal)
        # if self.code == "000A" and self.verb == " I":
        #     self._is_fragment = True if len(???.zones) > 8 else None
        # el
        if self.code == "0404" and self.verb == "RP":
            self._is_fragment = True
        elif self.code == "22C9" and self.verb == " I":
            self._is_fragment = None  # max length 24!
        else:
            self._is_fragment = False

        return self._is_fragment

    @property
    def is_valid(self) -> bool:  # Main code here
        """Parse the payload, return True if the message payload is valid.

        All exceptions are trapped, and logged appropriately.
        """

        if self._is_valid is not None:
            return self._is_valid

        try:  # determine which parser to use
            payload_parser = getattr(parsers, f"parser_{self.code}".lower())

        except AttributeError:  # there's no parser for this command code!
            payload_parser = getattr(parsers, "parser_unknown")

        try:  # run the parser
            self._payload = payload_parser(self.raw_payload, self)
            assert isinstance(self._payload, dict) or isinstance(self._payload, list)

        except (AssertionError, CorruptPayloadError):
            # beware: HGI80 can send parseable but 'odd' packets +/- get invalid reply
            if self.src.type == "18":  # TODO: should be a warning
                _LOGGER.warning(
                    "%s < Validation error (ignored)", self._pkt, extra=self.__dict__
                )
            else:
                _LOGGER.exception(
                    "%s < Validation error", self._pkt, extra=self.__dict__
                )
            self._is_valid = False
            return self._is_valid

        except (AttributeError, LookupError, TypeError, ValueError):  # for development
            _LOGGER.exception("%s < Coding error", self._pkt, extra=self.__dict__)
            self._is_valid = False
            return self._is_valid

        except NotImplementedError:  # unknown packet code
            _LOGGER.warning("%s < Unknown packet code", self._pkt, extra=self.__dict__)
            self._is_valid = False
            return self._is_valid

        else:
            self._is_valid = True

        # any remaining messages are valid, so: log them with one of these schemes
        if False and __dev_mode__:  # a hack to colourize by cycle
            if self.src.type == "01" and self.verb == " I":
                if (
                    self.code == "1F09"
                    or self.code in ("2309", "30C9", "000A")
                    and isinstance(self.payload, list)
                ):
                    _LOGGER.warning("%s", self, extra=self.__dict__)
                else:
                    _LOGGER.info("%s", self, extra=self.__dict__)
            else:
                _LOGGER.info("%s", self, extra=self.__dict__)

        elif False and __dev_mode__:  # a hack to colourize by verb
            if " I" in str(self):
                _LOGGER.info("%s", self, extra=self.__dict__)
            elif "RP" in str(self):
                _LOGGER.warning("%s", self, extra=self.__dict__)
            else:
                _LOGGER.error("%s", self, extra=self.__dict__)

        # elif self.src.id == "13:237335":
        elif self.verb == "RP" and self.src.type == "13":
            _LOGGER.warning("%s", self, extra=self.__dict__)

        elif self.code == "3B00":
            _LOGGER.info("%s", self, extra=self.__dict__)

        else:  # the normal mode logging scheme
            # TODO: parsing is 2x fast without this logging...
            _LOGGER.info("%s", self, extra=self.__dict__)
            pass

        return self._is_valid

    @exception_handler
    def create_devices(self) -> None:
        """Discover and create any new devices.

        Requires a valid packet; only 000C requires a valid message.
        """

        if self.code == "000C" and self.verb == "RP":
            if self.src.type == "01":  # TODO
                self._gwy.get_device(self.dst, controller=self.src)
                if self.is_valid:
                    key = "zone_idx" if "zone_idx" in self.payload else "domain_id"
                    [
                        self._gwy.get_device(
                            Address(id=d, type=d[:2]),
                            controller=self.src,
                            domain_id=self.payload[key],
                        )
                        for d in self.payload["devices"]
                    ]
            if self.src.type == "02":  # TODO
                if self.payload["devices"]:
                    device_id = self.payload["devices"][0]
                    self._gwy.get_device(
                        self.src, controller=Address(id=device_id, type=device_id[:2])
                    )

        elif self.src.type in ("01", "23", "30"):
            self._gwy.get_device(self.dst, controller=self.src)

        elif self.dst.type in ("01", "23", "30"):
            self._gwy.get_device(self.src, controller=self.dst)

        # TODO: will need other changes before these two will work...
        # TODO: the issue is, if the 1st pkt is not a 1F09 (or a list 000A/2309/30C9)
        # TODO: also could do 22D9 (UFC), others?
        # elif self.code == "1F09" and self.verb == " I":
        #     self._gwy.get_device(self.dst, controller=self.src)

        # elif self.code == "31D9" and self.verb == " I":  # HVAC
        #     self._gwy.get_device(self.dst, controller=self.src)
        # TODO: ...such as means to promote a device to a controller

        # this should catch all non-controller (and *some* controller) devices
        elif self.dst is self.src:
            self._gwy.get_device(self.src)

        # otherwise one will be a controller, *unless* dst is in ("--", "63")
        elif isinstance(self.src, Device) and self.src.is_controller:
            self._gwy.get_device(self.dst, controller=self.src)

        # TODO: may create a controller that doesn't exist
        elif isinstance(self.dst, Device) and self.dst.is_controller:
            self._gwy.get_device(self.src, controller=self.dst)

        else:
            self._gwy.get_device(self.src)
            self._gwy.get_device(self.dst)

        # where possible, swap each Address for its corresponding Device
        self.src = self._gwy.device_by_id.get(self.src.id, self.src)
        self.dst = self._gwy.device_by_id.get(self.dst.id, self.dst)

    @exception_handler
    def create_zones(self) -> None:
        """Discover and create any new zones (except HW)."""

        if not self.is_valid:  # requires self.payload
            return

        if self.src.type not in ("01", "23"):  # TODO: this is too restrictive!
            return

        # TODO: a I/0005: zones have changed & may need a restart (del) or not (add)
        if self.code == "0005":  # RP, and also I
            if self._payload["zone_type"] in CODE_0005_ZONE_TYPE.values():
                [
                    self.src.get_zone(
                        f"{idx:02X}",
                        zone_type=ZONE_TYPE_SLUGS.get(self._payload["zone_type"]),
                    )
                    for idx, flag in enumerate(self._payload["zone_mask"])
                    if flag == 1
                ]

        if self.code == "000C" and self.payload["devices"] and self.src.type == "01":
            devices = [self.src.device_by_id[d] for d in self.payload["devices"]]

            if self.payload["device_class"] == ATTR_ZONE_SENSOR:
                zone = self.src.get_zone(self.payload["zone_idx"])
                zone._set_sensor(devices[0])

            elif self.payload["device_class"] == "zone_actuators":
                # TODO: is this better, or...
                # self.src.get_zone(self.payload["zone_idx"], actuators=devices)
                # TODO: is it this one?
                zone = self.src.get_zone(self.payload["zone_idx"])
                for d in devices:
                    d.zone = zone

            elif self.payload["device_class"] == ATTR_HTG_CONTROL:
                self.src.boiler_control = devices[0]

            elif self.payload["device_class"] == ATTR_DHW_SENSOR:
                self.src.get_zone("FA")._set_sensor(devices[0])

            elif self.payload["device_class"] == ATTR_DHW_VALVE:
                self.src.get_zone("FA").hotwater_valve = devices[0]

            elif self.payload["device_class"] == ATTR_DHW_VALVE_HTG:
                self.src.get_zone("FA").heating_valve = devices[0]

        # # Eavesdropping (below) is used when discovery (above) is not an option
        # # TODO: needs work, e.g. RP/1F41 (excl. null_rp)
        # elif self.code in ("10A0", "1F41"):
        #     if isinstance(self.dst, Device) and self.dst.is_controller:
        #         self.dst.get_zone("FA")
        #     else:
        #         self.src.get_zone("FA")

        # # TODO: also process ufh_idx (but never domain_id)
        # elif isinstance(self._payload, dict):
        #     # TODO: only creating zones from arrays, presently, but could do so here
        #     if self._payload.get("zone_idx"):  # TODO: parent_zone too?
        #         if self.src.is_controller:
        #             self.src.get_zone(self._payload["zone_idx"])
        #         else:
        #             self.dst.get_zone(self._payload["zone_idx"])

        elif isinstance(self._payload, list):
            if self.code in ("000A", "2309", "30C9"):  # the sync_cycle pkts
                [self.src.get_zone(d["zone_idx"]) for d in self.payload]
            # elif self.code in ("22C9", "3150"):  # TODO: UFH zone
            #     pass

        # else:  # should never get here
        #     raise TypeError

    @exception_handler
    def update_entities(self, prev) -> None:  # TODO: needs work
        """Update the state of entities (devices, zones, ufh_zones)."""

        if not self.is_valid:  # requires self.payload
            return

        # HACK: merge 000A fragments
        # TODO: do here, or in ctl._proc_msg() and/or system._proc_msg()
        if re.search("I.* 01.* 000A ", self._pkt):  # HACK: and dtm < 3 secs
            # TODO: an edge case here: >2 000A packets in a row
            if prev is not None and re.search("I.* 01.* 000A ", prev._pkt):
                self._payload = prev.payload + self.payload  # merge frags, and process

        # some devices aren't created if they're filtered out
        if self.src.id not in self._gwy.device_by_id:
            return

        # some empty payloads may still be useful (e.g. RQ/3EF1/{})
        self._gwy.device_by_id[self.src.id]._proc_msg(self)

        # if payload is {} (empty dict; lists shouldn't ever be empty)
        if not self.payload:
            return

        # try to find the boiler relay, dhw sensor
        for evo in self._gwy.systems:
            if self.src.controller == evo.id:  # TODO: check!
                evo._proc_msg(self, prev)  # TODO: WIP
                if self.src.controller is not None:
                    break
            if self.src.id == evo.id:  # TODO: check!
                if self.code in ("10A0", "1260", "1F41") and evo._dhw is not None:
                    evo._dhw._proc_msg(self)

        # lists only useful to devices (c.f. 000C)
        if isinstance(self.payload, dict) and "zone_idx" in self.payload:
            evo = self.src.controller  # TODO: needs device?
            # if evo is None and isinstance(self.dst, Device):
            #     evo = self.dst._evo

            if evo is not None and self.payload["zone_idx"] in evo.zone_by_idx:
                evo.zone_by_idx[self.payload["zone_idx"]]._proc_msg(self)

            # elif self.payload.get("ufh_idx") in ...:  # TODO: is this needed?
            #     pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""The evohome-compatible system (is a 1-1 with a controller)."""

# import asyncio
from datetime import timedelta
import json
import logging
from typing import Optional

# from .command import Priority, RQ_RETRY_LIMIT, RQ_TIMEOUT
from .const import (
    ATTR_CONTROLLER,
    ATTR_DEVICES,
    ATTR_SYSTEM,
    # CODE_0005_ZONE_TYPE,
    # CODE_000C_DEVICE_TYPE,
    DEVICE_HAS_ZONE_SENSOR,
    DEVICE_TYPES,
    DISCOVER_SCHEMA,
    DISCOVER_PARAMS,
    DISCOVER_STATUS,
    DISCOVER_ALL,
    SYSTEM_MODE_LOOKUP,
    SYSTEM_MODE_MAP,
    __dev_mode__,
)
from .devices import _dtm, Controller, Device
from .exceptions import CorruptStateError
from .schema import (
    ATTR_HTG_CONTROL,
    ATTR_ORPHANS,
    ATTR_STORED_HOTWATER,
    ATTR_UFH_CONTROLLERS,
    ATTR_ZONES,
)

from .zones import DhwZone, Zone

_LOGGER = logging.getLogger(__name__)
if False and __dev_mode__:
    _LOGGER.setLevel(logging.DEBUG)
else:
    _LOGGER.setLevel(logging.WARNING)


class System(Controller):
    """The Controller base class, supports child devices and zones only."""

    def __init__(self, gateway, ctl_addr, **kwargs) -> None:
        super().__init__(gateway, ctl_addr, **kwargs)

        assert ctl_addr.id not in gateway.system_by_id, "Duplicate controller address"

        gateway.systems.append(self)
        gateway.system_by_id[self.id] = self

        self._dhw = None
        self._boiler_control = None

        self.zones = []
        self.zone_by_idx = {}
        # self.zone_by_name = {}

    def _proc_msg(self, msg) -> None:
        if msg.code in ("000A", "2309", "30C9") and not isinstance(msg.payload, list):
            pass
        else:
            super()._proc_msg(msg)

        if msg.code in ("000A", "2309", "30C9") and isinstance(msg.payload, list):
            pass

        # if msg.code in ("0005", "000C", "2E04"):
        #     pass
        # elif "zone_idx" in msg.payload:
        #     pass
        # else:
        #     assert False, "Unknown packet code"

    def __repr__(self) -> str:
        """Return an unambiguous string representation of this object."""
        return f"{self.id} ({DEVICE_TYPES.get(self.type)})"

    def __str__(self) -> str:  # TODO: WIP
        """Return a brief readable string representation of this object."""
        return json.dumps({self.id: self.schema})

    def get_zone(
        self, domain_id, zone_type=None, sensor=None, actuators=None
    ) -> Optional[Zone]:
        """Return a zone (will create it if required).

        Can also set a zone's sensor, and zone_type.
        """

        if domain_id == "FA":
            zone = self.dhw if self.dhw is not None else DhwZone(self)

        elif int(domain_id, 16) < self._gwy.config["max_zones"]:
            zone = self.zone_by_idx.get(domain_id)
            if zone is None:
                zone = Zone(self, domain_id)
            if zone_type is not None:
                zone._set_type(zone_type)

        elif domain_id in ("FC", "FF"):
            return

        else:
            raise ValueError("Unknown zone_type/domain_id")

        if sensor is not None:
            zone._set_sensor(sensor)  # TODO: check not an address

        if actuators is not None:
            zone.devices = actuators  # TODO: check not an address
            zone.device_by_id = {d.id: d for d in actuators}

        return zone

    @property
    def dhw(self) -> DhwZone:
        return self._dhw

    @dhw.setter
    def dhw(self, dhw: DhwZone) -> None:
        if not isinstance(dhw, DhwZone):
            raise ValueError

        if self._dhw is not None and self._dhw != dhw:
            raise CorruptStateError("The DHW has changed")

        if self._dhw is None:
            # self._gwy.get_device(xxx)
            # self.add_device(dhw.sensor)
            # self.add_device(dhw.relay)
            self._dhw = dhw

    @property
    def boiler_control(self) -> Device:
        return self._boiler_control

    @boiler_control.setter
    def boiler_control(self, device: Device) -> None:
        """Set the heater relay for this system (10: or 13:)."""

        if not isinstance(device, Device) or device.type not in ("10", "13"):
            raise TypeError

        if self._boiler_control is not None and self._boiler_control != device:
            raise CorruptStateError("The boiler relay has changed")
        # elif device.evo is not None and device.evo != self:
        #     raise LookupError  #  do this in self._gwy.get_device()

        if self._boiler_control is None:
            self._boiler_control = device
            device._set_domain(ctl=self)

    @property
    def schema(self) -> dict:
        """Return the system's schema."""

        schema = {ATTR_CONTROLLER: self.id}

        # devices without a parent zone, NB: CTL can be a sensor for a zones
        orphans = [d.id for d in self.devices if not d._domain_id and d.type != "02"]
        orphans.sort()
        # system" schema[ATTR_SYSTEM][ATTR_ORPHANS] = orphans

        schema[ATTR_SYSTEM] = {
            ATTR_HTG_CONTROL: self.boiler_control.id
            if self.boiler_control is not None
            else None,
            ATTR_ORPHANS: orphans,
        }

        schema[ATTR_STORED_HOTWATER] = self.dhw.schema if self.dhw is not None else None

        schema[ATTR_UFH_CONTROLLERS] = {
            u.id: u.schema
            for u in sorted(
                [d for d in self.devices if d.type == "02"], key=lambda x: x.id
            )
        }

        schema[ATTR_ZONES] = {
            z.idx: z.schema for z in sorted(self.zones, key=lambda x: x.idx)
        }

        schema["device_info"] = {
            d.id: d.hardware_info for d in self.devices if d.hardware_info is not None
        }

        return schema

    @property
    def params(self) -> dict:
        """Return the system's configuration."""

        params = {}

        params[ATTR_SYSTEM] = {
            "mode": self._get_msg_value("2E04"),  # **self.mode()
            "language": self._get_msg_value("0100", "language"),
            ATTR_HTG_CONTROL: {},
        }

        if self.boiler_control is not None:
            params[ATTR_SYSTEM][ATTR_HTG_CONTROL] = {
                "tpi_params": self.boiler_control._get_msg_value("1100"),
                "boiler_setpoint": self.boiler_control._get_msg_value("22D9"),
            }

        params[ATTR_STORED_HOTWATER] = self.dhw.params if self.dhw is not None else None

        params[ATTR_ZONES] = {
            z.idx: z.params for z in sorted(self.zones, key=lambda x: x.idx)
        }

        # ufh_controllers = [
        #     {d.id: d.config}
        #     for d in sorted(self.devices, key=lambda x: x.idx)
        #     if d.type == "02"
        # ]
        # ufh_controllers.sort()
        # config[ATTR_UFH_CONTROLLERS] = ufh_controllers

        # orphans = [
        #     {d.id: d.config}
        #     for d in sorted(self.devices, key=lambda x: x.idx)
        #     if d._zone is None
        #     # and d._ctl != d
        # ]  # devices without a parent zone, CTL can be a sensor for a zones
        # orphans.sort()
        # config[ATTR_ORPHANS] = orphans

        return params

    @property
    def status(self) -> dict:
        """Return the system's current state."""

        result = {}

        result[ATTR_SYSTEM] = {"datetime": self._get_msg_value("313F")}

        if self.boiler_control is not None:
            result[ATTR_SYSTEM][ATTR_HTG_CONTROL] = self.boiler_control.status

        result[ATTR_STORED_HOTWATER] = self.dhw.status if self.dhw is not None else None

        result[ATTR_ZONES] = {
            z.idx: z.status for z in sorted(self.zones, key=lambda x: x.idx)
        }

        result[ATTR_DEVICES] = {
            d.id: d.status
            for d in sorted(self.devices, key=lambda x: x.id)
            if d.id != self.id
        }

        return result

    @property
    def tpi_params(self) -> Optional[float]:  # 1100
        return self._get_msg_value("1100")

    @property
    def sync_tpi(self) -> Optional[float]:  # 3B00
        return self._get_msg_value("3B00", "sync_tpi")


class EvoSystem(System):
    """The EvoSystem class - some controllers are evohome-compatible."""

    def __init__(self, gateway, ctl_addr, **kwargs) -> None:
        super().__init__(gateway, ctl_addr, **kwargs)

        self._prev_30c9 = None
        self._fault_log = {}

        # self._discover()

    def _discover(self, discover_flag=DISCOVER_ALL) -> None:
        if self._gwy.config["disable_discovery"]:
            return
        super()._discover()

        if discover_flag & DISCOVER_SCHEMA:
            [  # 000C: find the HTG relay and DHW sensor, if any (DHW relays in DHW)
                self._send_cmd("000C", payload=dev_type)
                for dev_type in ("000F", "000D")  # CODE_000C_DEVICE_TYPE
                # for dev_type, description in CODE_000C_DEVICE_TYPE.items() fix payload
                # if description is not None
            ]

            [  # 0005: find any configured zones, + their type (RAD, UFH, VAL, MIX, ELE)
                self._send_cmd("0005", payload=f"00{zone_type}")
                for zone_type in ("08", "09", "0A", "0B", "11")  # CODE_0005_ZONE_TYPE
                # for zone_type, description in CODE_0005_ZONE_TYPE.items()
                # if description is not None
            ]

        if discover_flag & DISCOVER_PARAMS:
            self._send_cmd("1100", payload="FC")  # TPI params
            self._send_cmd("0100")  # language

        if discover_flag & DISCOVER_STATUS:
            self._send_cmd("2E04", payload="FF")  # system mode
            for code in ("1F09", "313F"):  # system_sync, datetime
                self._send_cmd(code)

        # TODO: test only
        # asyncio.create_task(
        #     self.async_set_mode(5, dt_now() + timedelta(minutes=120))
        #     # self.async_set_mode(5)
        #     # self.async_reset_mode()
        # )

        # # for code in ("3B00"):  # 3EF0, 3EF1
        # #     for payload in ("0000", "00", "F8", "F9", "FA", "FB", "FC", "FF"):
        # #         self._send_cmd(code, payload=payload)

        # # TODO: opentherm: 1FD4, 22D9, 3220

        # TODO: Get the fault log entries
        # self._fault_log.req_log(log_idx=0)
        # # for log_idx in range(0, 0x6):  # max is 0x3C?, 0x3F (highest log is 0x3E?)
        # #     self._send_cmd("0418", payload=f"{log_idx:06X}", priority=Priority.LOW)

    def _proc_msg(self, msg, prev_msg=None):
        """Eavesdrop packets, or pairs of packets, to maintain the system state."""

        def is_exchange(this, prev):  # TODO:use is?
            return this.src is prev.dst and this.dst is prev.src.addr

        def find_htg_relay(this, prev=None):
            """Discover the heat relay (10: or 13:) for this system.

            There's' 3 ways to find a controller's heat relay (in order of reliability):
            1.  The 3220 RQ/RP *to/from a 10:* (1x/5min)
            2a. The 3EF0 RQ/RP *to/from a 10:* (1x/1min)
            2b. The 3EF0 RQ (no RP) *to a 13:* (3x/60min)
            3.  The 3B00 I/I exchange between a CTL & a 13: (TPI cycle rate, usu. 6x/hr)

            Data from the CTL is considered 'authorative'. The 1FC9 RQ/RP exchange
            to/from a CTL is too rare to be useful.
            """

            # 18:14:14.025 066 RQ --- 01:078710 10:067219 --:------ 3220 005 0000050000
            # 18:14:14.446 065 RP --- 10:067219 01:078710 --:------ 3220 005 00C00500FF
            # 14:41:46.599 064 RQ --- 01:078710 10:067219 --:------ 3EF0 001 00
            # 14:41:46.631 063 RP --- 10:067219 01:078710 --:------ 3EF0 006 0000100000FF  # noqa

            # 06:49:03.465 045 RQ --- 01:145038 13:237335 --:------ 3EF0 001 00
            # 06:49:05.467 045 RQ --- 01:145038 13:237335 --:------ 3EF0 001 00
            # 06:49:07.468 045 RQ --- 01:145038 13:237335 --:------ 3EF0 001 00
            # 09:03:59.693 051  I --- 13:237335 --:------ 13:237335 3B00 002 00C8
            # 09:04:02.667 045  I --- 01:145038 --:------ 01:145038 3B00 002 FCC8

            # note the order: most to least reliable
            heater = None

            if this.code == "3220" and this.verb == "RQ":
                if this.src is self and this.dst.type == "10":
                    heater = this.dst

            elif this.code == "3EF0" and this.verb == "RQ":
                if this.src is self and this.dst.type in ("10", "13"):
                    heater = this.dst

            elif this.code == "3B00" and this.verb == " I" and prev is not None:
                if prev.code == this.code and prev.verb == this.verb:
                    if this.src is self and prev.src.type == "13":
                        heater = prev.src

            if heater is not None:
                self.boiler_control = heater

        def find_dhw_sensor(this):
            """Discover the stored HW this system (if any).

            There is only 2 way2 to find a controller's DHW sensor:
            1. The 10A0 RQ/RP *from/to a 07:* (1x/4h) - reliable
            2. Use sensor temp matching - non-deterministic

            Data from the CTL is considered more authorative. The RQ is initiated by the
            DHW, so is not authorative. The I/1260 is not to/from a controller, so is
            not useful.
            """

            # 10A0: RQ/07/01, RP/01/07: can get both parent controller & DHW sensor
            # 047 RQ --- 07:030741 01:102458 --:------ 10A0 006 00181F0003E4
            # 062 RP --- 01:102458 07:030741 --:------ 10A0 006 0018380003E8

            # 1260: I/07: can't get which parent controller - need to match temps
            # 045  I --- 07:045960 --:------ 07:045960 1260 003 000911

            # 1F41: I/01: get parent controller, but not DHW sensor
            # 045  I --- 01:145038 --:------ 01:145038 1F41 012 000004FFFFFF1E060E0507E4
            # 045  I --- 01:145038 --:------ 01:145038 1F41 006 000002FFFFFF

            sensor = None

            if this.code == "10A0" and this.verb == "RP":
                if this.src is self and this.dst.type == "07":
                    sensor = this.dst

            if sensor is not None:
                if self.dhw is None:
                    self.get_zone("FA")
                self.dhw._set_sensor(sensor)

        def find_zone_sensors() -> None:
            """Determine each zone's sensor by matching zone/sensor temperatures.

            The temperature of each zone is reliably known (30C9 array), but the sensor
            for each zone is not. In particular, the controller may be a sensor for a
            zone, but unfortunately it does not announce its sensor temperatures.

            In addition, there may be 'orphan' (e.g. from a neighbour) sensors
            announcing temperatures with the same value.

            This leaves only a process of exclusion as a means to determine which zone
            uses the controller as a sensor.
            """

            # A reasonable assumption from this point on: a zone's _temperature attr has
            # just been updated via the controller's 30C9 pkt, and hasn't changed since.
            # It's also assumed that the gateway (18:) has received the same 30C9 pkts
            # from the sensors as the controller has: for some this may not be reliable.
            # The final assumption: the controller, as a sensor, has a temp distinct
            # from all others (so another sensor isn't matched to the controllers zone).
            # If required (and it's not clear that it is required), the above can be
            # mitigated by confirming a sensor after two (consistent) matches.

            prev_msg, self._prev_30c9 = self._prev_30c9, msg
            if prev_msg is None:
                return

            if len([z for z in self.zones if z.sensor is None]) == 0:
                return  # (currently) no zone without a sensor

            # if self._gwy.serial_port:  # only if in monitor mode...
            secs = self._get_msg_value("1F09", "remaining_seconds")
            if secs is None or msg.dtm > prev_msg.dtm + timedelta(seconds=secs):
                return  # only compare against 30C9 (array) pkt from the last cycle

            _LOGGER.debug("System state (before): %s", self)

            changed_zones = {
                z["zone_idx"]: z["temperature"]
                for z in msg.payload
                if z not in prev_msg.payload
            }  # zones with changed temps
            _LOGGER.debug("Changed zones (from 30C9): %s", changed_zones)
            if not changed_zones:
                return  # ctl's 30C9 says no zones have changed temps during this cycle

            testable_zones = {
                z: t
                for z, t in changed_zones.items()
                if self.zone_by_idx[z].sensor is None
                and t not in [v for k, v in changed_zones.items() if k != z] + [None]
            }  # ...with unique (non-null) temps, and no sensor
            _LOGGER.debug(
                " - with unique/non-null temps (from 30C9), no sensor (from state): %s",
                testable_zones,
            )
            if not testable_zones:
                return  # no testable zones

            testable_sensors = [
                d
                for d in self._gwy.devices  # not: self.devices
                if d._ctl in (self, None)
                and d.addr.type in DEVICE_HAS_ZONE_SENSOR
                and d.temperature is not None
                and d._msgs["30C9"].dtm > prev_msg.dtm  # changed temp during last cycle
            ]

            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug(
                    "Testable zones: %s (unique/non-null temps & sensorless)",
                    testable_zones,
                )
                _LOGGER.debug(
                    "Testable sensors: %s (non-null temps & orphans or zoneless)",
                    {d.id: d.temperature for d in testable_sensors},
                )

            if testable_sensors:  # the main matching algorithm...
                for zone_idx, temp in testable_zones.items():
                    # TODO: when sensors announce temp, ?also includes it's parent zone
                    matching_sensors = [
                        s
                        for s in testable_sensors
                        if s.temperature == temp and s._zone in (zone_idx, None)
                    ]
                    _LOGGER.debug("Testing zone %s, temp: %s", zone_idx, temp)
                    _LOGGER.debug(
                        " - matching sensor(s): %s (same temp & not from another zone)",
                        [s.id for s in matching_sensors],
                    )

                    if len(matching_sensors) == 1:
                        _LOGGER.debug("   - matched sensor: %s", matching_sensors[0].id)
                        zone = self.zone_by_idx[zone_idx]
                        zone._set_sensor(matching_sensors[0])
                        zone.sensor.controller = self
                    elif len(matching_sensors) == 0:
                        _LOGGER.debug("   - no matching sensor (uses CTL?)")
                    else:
                        _LOGGER.debug("   - multiple sensors: %s", matching_sensors)

                _LOGGER.debug("System state (after): %s", self)

            # now see if we can allocate the controller as a sensor...
            if self._zone is not None:
                return  # the controller has already been allocated
            if len([z for z in self.zones if z.sensor is None]) != 1:
                return  # no single zone without a sensor

            testable_zones = {
                z: t
                for z, t in changed_zones.items()
                if self.zone_by_idx[z].sensor is None
            }  # this will be true if ctl is sensor
            if not testable_zones:
                return  # no testable zones

            zone_idx, temp = list(testable_zones.items())[0]
            _LOGGER.debug("Testing (sole remaining) zone %s, temp: %s", zone_idx, temp)
            # want to avoid complexity of z._temperature
            # zone = self.zone_by_idx[zone_idx]
            # if zone._temperature is None:
            #     return  # TODO: should have a (not-None) temperature

            matching_sensors = [
                s
                for s in testable_sensors
                if s.temperature == temp and s._zone in (zone_idx, None)
            ]

            _LOGGER.debug(
                " - matching sensor(s): %s (excl. controller)",
                [s.id for s in matching_sensors],
            )

            # can safely(?) assume this zone is using the CTL as a sensor...
            if len(matching_sensors) == 0:
                _LOGGER.debug("   - matched sensor: %s (by exclusion)", self.id)
                zone = self.zone_by_idx[zone_idx]
                zone._set_sensor(self)
                zone.sensor.controller = self

            _LOGGER.debug("System state (finally): %s", self)

        super()._proc_msg(msg)

        # if msg.code == "0005" and prev_msg is not None:
        #     zone_added = bool(prev_msg.code == "0004")  # else zone_deleted

        if msg.code == "0418" and msg.verb in (" I", "RP"):  # this is a special case
            _LOGGER.debug("Zone(%s).update: Received RP/0418 (fault_log)", self.id)
            # self._fault_log.add_entry(msg)
            # do the following only if we had: self._fault_log.req_log(log_idx=0)
            # self._fault_log.req_entry(log_idx=payload["log_idx"] + 1)
            if "log_idx" in msg.payload:
                self._fault_log[msg.payload["log_idx"]] = msg

        if msg.code == "30C9" and isinstance(msg.payload, list):  # msg.is_array:
            find_zone_sensors()

        if msg.code == "3EF1" and msg.verb == "RQ":  # relay attached to a burner
            if msg.dst.type == "13":  # this is the TPI relay
                pass
            if msg.dst.type == "10":  # this is the OTB
                pass

        if prev_msg is not None and prev_msg.src.controller is not None:
            if prev_msg.src.controller is not self:
                return

        # if msg.src.type == "01" and msg.dst.controller is None:  # 3EF0
        #     msg.dst.controller = msg.src  # useful for TPI/OTB, uses 3EF0

        if msg.code in ("3220", "3B00", "3EF0"):  # self.boiler_control is None and
            find_htg_relay(msg, prev=prev_msg)

        if msg.code in ("10A0", "1260"):  # self.dhw.sensor is None and
            find_dhw_sensor(msg)

        # else:
        #     assert False, "Unknown packet code"

    # def fault_log(self, force_update=False) -> Optional[list]:  # 0418
    #     # TODO: try to discover fault codes
    #     for log_idx in range(0x00, 0x3C):  # 10 pages of 6
    #         self._send_cmd("0418", payload=f"{log_idx:06X}")

    #     return [f.payload for f in self._fault_log.values()]

    @property
    def language(self) -> Optional[str]:  # 0100,
        return self._get_msg_value("0100", "language")

    @property
    def mode(self) -> dict:  # 2E04
        """Return the system mode."""
        return self._mode

    async def set_mode(self, mode, until=None):  # 2E04
        """Set the system mode for a specified duration, or indefinitely."""

        if isinstance(mode, int):
            mode = f"{mode:02X}"
        elif not isinstance(mode, str):
            raise TypeError("Invalid system mode")
        elif mode in SYSTEM_MODE_LOOKUP:
            mode = SYSTEM_MODE_LOOKUP[mode]

        if mode not in SYSTEM_MODE_MAP:
            raise ValueError("Unknown system mode")

        until = _dtm(until) + "00" if until is None else "01"

        self._send_cmd("2E04", verb=" W", payload=f"{mode}{until}")

    async def reset_mode(self) -> None:  # 2E04
        """Revert the system mode to Auto."""  # TODO: is it AutoWithReset?
        self._send_cmd("2E04", verb=" W", payload="00FFFFFFFFFFFF00")
